"""# Standardize test format

Tool for standardizing Partisia's Java tests.

Is capable of migrating from Javadoc based property specification to the
`@DisplayName` specification and back.

## Usage

Convert tests with `@DisplayName` to Javadoc:

```bash
python -m standardize_java_text --javadoc --only-displayname FILE...
```

It it recommended that you do this with no unstaged changes, to more easily
verify the result.

You may need to call your formatting tool afterwards, as the converter doesn't
take formatting into account.
"""

import enum
import hashlib
import re
import dataclasses

from ._version import __version__

__all__ = ['__version__', 'standardize_java_text', 'NamingScheme', 'Config']



class NamingScheme(enum.Enum):
    PRESERVE = 'preserve'
    HASH_OF_DESC = 'hash'
    FROM_DESC = 'desc'


@dataclasses.dataclass
class Config:
    with_javadoc: bool
    with_display_name: bool
    only_for_javadoc: bool
    only_for_display_name: bool
    naming_scheme: NamingScheme


PATTERN_JAVADOC = r'/\*\*(?P<javadoc>.*(\n\s*\*.*)*?)\s*\*/'
PATTERN_DISPLAY_NAME_EXPRESSION = r'\s*(?:"[^"]*?")(?:\s*\+\s*"[^"]*?")*\s*'
PATTERN_DISPLAY_NAME = (
    r'@DisplayName\((?P<displayname>' + PATTERN_DISPLAY_NAME_EXPRESSION + r')\)'
)

TEST_PATTERN: re.Pattern = re.compile(
    r''
    + r'(?:'
    + PATTERN_JAVADOC
    + r'\s*)?'
    + r'@(?P<annotation>Test|BeforeEach|ParameterizedTest)\s*'
    + r'(?:'
    + PATTERN_DISPLAY_NAME
    + r'\s*)?'
    + r'(?P<visibility>public\s+|private\s+)?void\s+(?P<name>\w+)\(',
    re.IGNORECASE,
)

END_SYMBOLS = '.?!'

IGNORABLE_WORDS_IN_NAME = frozenset(
    {
        '',
        'a',
        'an',
        'the',
        'is',
        'are',
        'test',
        'tests',
    },
)


def format_test_prefix(
    name: str,
    description: str,
    annotation: str,
    visibility: str,
    with_javadoc: bool,
    with_display_name: bool,
) -> str:
    indent = ''
    str_builder = []
    if with_javadoc:
        str_builder += ['/** ', description, ' */', '\n']
    str_builder += [indent, '@', annotation]
    if with_display_name:
        str_builder += ['\n', indent, '@DisplayName("']
        str_builder.append(description.rstrip(END_SYMBOLS))
        str_builder += ['")']
    str_builder += ['\n', indent]
    if visibility:
        str_builder += [visibility, ' ']
    str_builder += ['void ', name, '(']
    return ''.join(str_builder)


def to_camel_case_word(word: str, first: bool = False) -> str | None:
    if len(word) > 1 and word.upper() == word:
        word = word.lower()
    return (word[0].lower() if first else word[0].upper()) + word[1:]


def to_camel_case(description: str) -> str:
    description = (
        description.replace(',', ' ')
        .replace('.', ' ')
        .replace('"', ' ')
        .replace('+', ' ')
        .replace('-', ' ')
        .replace(':', ' ')
        .replace("'", '')
        .strip(' \t.')
    )
    words = description.split(' ')

    words = [w.strip() for w in words]
    words = [w for w in words if w.lower() not in IGNORABLE_WORDS_IN_NAME]

    words[0] = to_camel_case_word(words[0], first=True)
    for i in range(1, len(words)):
        words[i] = to_camel_case_word(words[i])

    words = [w for w in words if w]
    return ''.join(words)


def from_camel_case(name: str) -> str:
    return re.sub(
        r'(^|[A-Z])[a-z]*',
        lambda x: ' ' + x.group(0).capitalize(),
        name,
    ).strip()


def parse_javadoc_to_description(text: str | None):
    if text is None:
        return ''
    text = re.sub(r'\n\s*\*', ' ', text).strip()
    text = re.sub(r'\{@link\s+(\w+)#(\w+)\}', r'\1 \2', text)
    text = re.sub(r'\{@link\s+(\w+)\}', r'\1', text)
    text = re.sub(r'<p>', r'', text)
    text = re.sub(r'</?b>', r'', text)
    return text


def parse_display_name_to_description(text: str | None):
    if text is None:
        return ''

    text = text.strip()
    text = re.sub(r'"\s*\+\s*"', '', text)
    text = text.strip('"')
    text = text.strip()
    return text


def replace_test_pattern(
    match: re.Match,
    config: Config,
) -> str:
    javadoc = parse_javadoc_to_description(match.group('javadoc'))

    annotation = match.group('annotation').strip()
    visibility = (match.group('visibility') or '').strip()
    name = match.group('name').strip()
    displayname = parse_display_name_to_description(match.group('displayname'))

    # Filtering
    if config.only_for_javadoc and not javadoc:
        return match.group(0)
    if config.only_for_display_name and not displayname:
        return match.group(0)

    # Formatting
    description = displayname or javadoc or from_camel_case(name)
    description = re.sub(r'[ \t]+', ' ', description)

    if config.naming_scheme == NamingScheme.PRESERVE:
        pass
    elif config.naming_scheme == NamingScheme.FROM_DESC:
        name = to_camel_case(description)
    elif config.naming_scheme == NamingScheme.HASH_OF_DESC:
        h = hashlib.sha256()
        h.update(description.encode('utf8'))
        name = 'test' + h.hexdigest()
        del h

    return format_test_prefix(
        name,
        description,
        annotation,
        visibility,
        with_javadoc=config.with_javadoc,
        with_display_name=config.with_display_name,
    )


IMPORT_DISPLAY_NAME = 'import org.junit.jupiter.api.DisplayName;'


def standardize_java_text(
    text: str,
    config: Config,
):
    text = TEST_PATTERN.sub(
        lambda m: replace_test_pattern(
            m,
            config,
        ),
        text,
    )
    if '@DisplayName' in text and IMPORT_DISPLAY_NAME not in text:
        lines = text.split('\n')
        lines.insert(1, IMPORT_DISPLAY_NAME)
        text = '\n'.join(lines)
        del lines
    return text
